#!/usr/bin/env python
# $Id: server.py,v 1.103 2013/05/14 15:53:25 irees Exp $
import traceback
import thread
import os.path
import functools
import contextlib
import time
import collections
import sys

# Twisted imports
from twisted.application import internet
import twisted.internet
import twisted.internet.reactor
import twisted.web.static
import twisted.web.server
import twisted.python.threadpool

try:
    from twisted.internet import ssl
except ImportError:
    ssl = None

import emen2.db.config
import emen2.db.log
import emen2.web.routing

class DBPool(object):
    """Simple DB Pool loosely based on twisted.enterprise.adbapi.ConnectionPool."""
    # All connections, key is Thread ID
    dbs = {}
    running = False

    def __init__(self, min=0, max=5):
        # Minimum and maximum number of threads
        self.min = min
        self.max = max
        # Generate Thread ID
        self.threadID = thread.get_ident
        # Connect to reactor
        self.reactor = twisted.internet.reactor
        self.threadpool = twisted.python.threadpool.ThreadPool(self.min, self.max)

    def start(self):
        """Start a database connection on startup to run recovery, setup, etc."""
        self.connect()

    def connect(self):
        """Create a new database connection."""
        import emen2.db.database
        tid = self.threadID()
        # emen2.db.log.info('DBPool info: # threads: %s -- this thread is %s'%(len(self.dbs), tid))
        db = self.dbs.get(tid)
        if not db:
            db = emen2.db.database.opendb()
            self.dbs[tid] = db
        return db

    def disconnect(self, db):
        """Disconnect a database connection."""
        tid = self.threadID()
        if db is not self.dbs.get(tid):
            raise Exception('Wrong connection for thread')
        if db:
            # db.close()
            del self.dbs[tid]

    def rundb(self, call, *args, **kwargs):
        return twisted.internet.threads.deferToThread(self._rundb, call, *args, **kwargs)

    def _rundb(self, call, *args, **kwargs):
        db = self.connect()
        kwargs['db'] = db
        result = call(*args, **kwargs)
        return result


# The thread pool.
pool = DBPool()


##### Web server ######

class WebServerOptions(emen2.db.config.DBOptions):
    optParameters = [
        ['port',      'HTTP port',  None, None, int],
        ['httpsport', 'HTTPS port', None, None, int],
    ]

    optFlags = [
        ['https', None, 'Use HTTPS']
    ]

class EMEN2Site(twisted.web.server.Site):
    def log(self, request):
        # rfc identd used for client supplied session ID
        # userid field is authenticated user name
        # This is a hack to get the session/username
        ctxid = getattr(request, "_log_ctxid", None) or "-"
        username = getattr(request, "_log_username", None) or "-"

        line = '%s %s %s %s "%s" %d %s "%s" "%s"' % (
            request.getClientIP(),
            ctxid,
            username,
            self._logDateTime,
            '%s %s %s' % (self._escape(request.method),
                          self._escape(request.uri),
                          self._escape(request.clientproto)),
            request.code,
            request.sentLength or "-",
            self._escape(request.getHeader("referer") or "-"),
            self._escape(request.getHeader("user-agent") or "-"))

        emen2.db.log.web(line)

class EMEN2BaseServer(object):

    usage = WebServerOptions

    def __init__(self, options=None):
        options = options or {}
        self.port = options.get('port') or emen2.db.config.get('web.port')

    #@contextlib.contextmanager
    def start(self, service=None):
        '''Run the server main loop'''

        pool.start()

        # Routing resource. This will look up request.uri in the routing table
        # and return View resources.
        root = emen2.web.routing.Router()

        # Previously this used contextmanager and yield to attach the resources.
        self.attach_resources(root)

        # The Twisted Web server protocol factory,
        #  with our Routing resource as root
        # self.site = twisted.web.server.Site(root)
        self.site = EMEN2Site(root)

        reactor = twisted.internet.reactor
        reactor.suggestThreadPoolSize(emen2.db.config.get('web.threads', 1))

        # Attach to a service, or run standalone.
        if service:
            self.attach_to_service(service)
        else:
            self.attach_standalone()

    def attach_resources(self, root):
        pass
        
    def attach_to_service(self, service):
        emen2_service = internet.TCPServer(self.port, self.site)
        emen2_service.setServiceParent(service)

    def attach_standalone(self):
        reactor = twisted.internet.reactor
        reactor.listenTCP(self.port, self.site)
        reactor.run()

class EMEN2RPCServer(EMEN2BaseServer):
    """Only start the RPC server."""
    def attach_resources(self, root):
        import jsonrpc.server
        from emen2.web.resource import JSONRPCServerEvents
        root.putChild('jsonrpc', jsonrpc.server.JSON_RPC().customize(JSONRPCServerEvents))
    
class EMEN2WebServer(EMEN2BaseServer):
    """Start the full web server."""
    def attach_resources(self, root):
        # Load all View extensions
        import emen2.db.config
        emen2.db.config.load_views()

        # Child resources that do not go through the Router.
        import jsonrpc.server
        import emen2.web.resource

        # if all the JSON_RPC class does is change the eventhandler, it can (and should) be instantiated this way:
        from emen2.web.resource import JSONRPCServerEvents
        root.putChild('jsonrpc', jsonrpc.server.JSON_RPC().customize(JSONRPCServerEvents))
        root.putChild('static', twisted.web.static.File(emen2.db.config.get_filename('emen2', 'web/static')))
        root.putChild('static-%s'%emen2.__version__, twisted.web.static.File(emen2.db.config.get_filename('emen2', 'web/static')))
        root.putChild('favicon.ico', twisted.web.static.File(emen2.db.config.get_filename('emen2', 'web/static/favicon.ico')))
        root.putChild('robots.txt', twisted.web.static.File(emen2.db.config.get_filename('emen2', 'web/static/robots.txt')))
    
def start_standalone():
    opt = emen2.db.config.UsageParser(WebServerOptions)
    server = EMEN2WebServer(opt.options)
    server.start()
    
def start_rpc():
    opt = emen2.db.config.UsageParser(WebServerOptions)
    server = EMEN2RPCServer(opt.options)
    server.start()

if __name__ == "__main__":
    start_standalone()

__version__ = "$Revision: 1.103 $".split(":")[1][:-1].strip()


# $Id: group.py,v 1.16 2013/05/01 08:22:10 irees Exp $
import time

from emen2.web.view import View


@View.register
class Groups(View):

    @View.add_matcher(r'^/groups/$')
    def main(self,q=None):
        self.template="/pages/groups"
        self.title = "User group directory"
        self.ctxt["q"] = ""
        groupnames = self.db.group.filter(None)
        groups = self.db.group.get(groupnames)
        admin = self.db.auth.check.admin()
        self.ctxt["admin"] = admin

        if groups == None:
            self.simple(content="""No user groups found, or insufficient permissions to view user group list.""")
            return
        
        self.ctxt['groupnames'] = groupnames
        self.ctxt['groups'] = groups



@View.register
class Group(View):
    
    @View.add_matcher(r'^/group/(?P<name>[^/]*)/$')
    def main(self, name=None):
        group = self.db.group.get(name)
        self.title = "User group: %s"%(group.displayname)
        self.template = "/pages/group"
        self.ctxt['group'] = group
        self.ctxt['new'] = False
        self.ctxt['edit'] = False


    @View.add_matcher(r'^/group/(?P<name>[^/]*)/edit/$')
    def edit(self, name=None, **kwargs):
        group = self.db.group.get(name)
        self.title = "User group: %s"%(group.displayname)
        self.template = "/pages/group"
        self.ctxt['group'] = group
        self.ctxt['new'] = False
        self.ctxt['edit'] = True

        if self.request_method != 'post':
            return

        group.update(kwargs)
        group = self.db.group.put(group)
        self.ctxt['group'] = group
        self.redirect('/group/%s/'%group.name)
        

    @View.add_matcher(r'^/groups/new/$')
    def new(self, name=None, **kwargs):
        # We have to supply a group name.. just use a random string.
        group = self.db.group.new(name=name)        
        self.ctxt['group'] = group
        self.ctxt['new'] = True
        self.ctxt['edit'] = True
        self.title = "New group"
        self.template = "/pages/group"

        if self.request_method != 'post':
            return
            
        group.update(kwargs)
        group = self.db.group.put(group)
        self.ctxt['group'] = group
        self.redirect('/group/%s/'%group.name)
        
        

        
__version__ = "$Revision: 1.16 $".split(":")[1][:-1].strip()



# $Id: auth.py,v 1.20 2012/10/19 10:12:45 irees Exp $
from emen2.web.view import View

@View.register
class Auth(View):

    @View.add_matcher(r'^/auth/login/$')
    def login(self, username=None, password=None, msg='', errmsg='', **kwargs):
        self.template = '/auth/login'
        self.title = 'Login'
        self.ctxt["username"] = username
        if username != None:
            ctxid = self.db.auth.login(username, password, host=self.request_host)
            self.notify('Successfully logged in.')
            self.set_header('X-Ctxid', ctxid)
            self.redirect('/')

    @View.add_matcher(r'^/auth/logout/$')
    def logout(self, msg='', **kwargs):
        self.template = '/auth/login'
        self.title = 'Logout'
        self.db.auth.logout()
        self.set_header('X-Ctxid', '')
        self.redirect(content='Successfully logged out.', auto=False)

    @View.add_matcher(r'^/auth/password/change/$', name='password/change')
    def setpassword(self, **kwargs):
        self.template = '/auth/password.change'
        self.title = "Password change"

        name = kwargs.pop("name",None) or self.db.auth.check.context()[0]
        opw = kwargs.pop("opw",None)
        on1 = kwargs.pop("on1",None)
        on2 = kwargs.pop("on2",None)

        self.ctxt['name'] = name
        if not on1 and not on2:
            pass

        elif on1 != on2:
            self.notify("New passwords did not match", error=True)

        else:
            try:
                self.db.user.setpassword(name, opw, on1)
                self.redirect(content="Password changed successfully.", auto=False)
            except Exception, errmsg:
                self.notify(errmsg, error=True)

    @View.add_matcher(r'^/auth/password/reset/$', name='password/reset')
    @View.add_matcher(r'^/auth/password/reset/(?P<name>[^/]*)/(?P<secret>\w+)/$', name='password/reset/confirm')
    def resetpassword(self, email=None, name=None, secret=None, newpassword=None, **kwargs):
        self.template = '/auth/password.reset'
        self.title = "Password reset"
        self.ctxt['name'] = name
        self.ctxt['email'] = email
        self.ctxt['secret'] = secret
        self.ctxt['newpassword'] = ''
        if name and secret and newpassword:
            try:
                self.db.user.setpassword(name, oldpassword=None, newpassword=newpassword, secret=secret)
                self.redirect(content='The password for your account has been changed.', auto=False)
            except Exception, errmsg:
                self.notify(errmsg, error=True)

        elif email:
            try:
                self.db.user.resetpassword(email)
                self.redirect(content='Instructions for resetting your password have been sent to %s.'%email, auto=False)
            except Exception, errmsg:
                self.notify(errmsg, error=True)


    @View.add_matcher(r'^/auth/email/change/$', name='email/change')
    def setemail(self, **kwargs):
        self.template = '/auth/email.change'
        self.title = "Change email"

        name = kwargs.get("name") or self.db.auth.check.context()[0]
        opw = kwargs.get('opw', '')
        email = kwargs.get('email', '')

        self.ctxt['email'] = email

        if email:
            try:
                user = self.db.user.setemail(name=name, email=email, password=opw)
                if email == user.email:
                    self.redirect(content='Email address successfully updated to %s.'%user.email, auto=False)
                else:
                    self.redirect(content='A verification email has been sent to %s.'%email, auto=False)
            except Exception, errmsg:
                self.notify(errmsg, error=True)
                

    @View.add_matcher(r'^/auth/email/verify/(?P<name>[^/]*)/(?P<email>[^/]*)/(?P<secret>\w+)/$', name='email/verify')
    def verifyemail(self, name=None, email=None, secret=None, **kwargs):
        self.template = '/auth/email.verify'
        self.title = "Verify email"

        if name and secret:
            try:
                user = self.db.user.setemail(name=name, email=email, secret=secret)
                self.redirect(content="The email address for your account has been changed to %s"%user.email, auto=False)
            except Exception, errmsg:
                self.notify(errmsg, error=True)


__version__ = "$Revision: 1.20 $".split(":")[1][:-1].strip()



# $Id: recorddef.py,v 1.42 2013/05/01 03:16:07 irees Exp $
"""RecordDef (Protocol) classes

Classes:
    RecordDef
    RecordDefDB

"""

import re
import textwrap

# EMEN2 imports
import emen2.db.dataobject


def parseparmvalues(text):
    regex = emen2.db.database.VIEW_REGEX

    params = set()
    required = set()
    defaults = {}

    for match in regex.finditer(text):
        n = match.group('name')
        t = match.group('type')

        if not n:
            continue
        if t == '$' or t == '*':
            params.add(match.group('name'))
            if match.group('def'):
                defaults[n] = match.group('def')
        if t == '*':
            required.add(n)

    return params, defaults, required



class RecordDef(emen2.db.dataobject.BaseDBObject):
    """RecordDefs, aka Protocols.

    RecordDefs, aka Protocols, function as templates for Records. Each Record must be an instance of a defined RecordDef.
    The RecordDef defines the default parameters that make up a record, and a set of presentation formats ('views'). The 'mainview'
    is parsed for parameters and becomes the default view. Other important views that are used by the web interface are:

        recname            A simple title for each record built from record values
        tabularview        Columns to use in table views

    RecordDefs may have parent/child relationships, similar to Records.

    @attr desc_short Short description. Shown in many places in the interface as the label for this RecordDef.
    @attr desc_long Long description. Shown as help in new record page.
    @attr mainview Default protocol view. This should be an expanded form, similar to a lab notebook or experimental protocol. This should be IMMUTABLE, as it may describe a particular experiment. However, admins are allowed to edit the mainview if it is absolutely necessary.
    @attr views Dictionary of additional views. Usually recname/tabularview will be defined, as well as defaultview
    @attr private Mark this RecordDef as private. False for public access, True for private. If private, you must be admin or able to read a record of this type to access the RecordDef
    @attr typicalchld A list of RecordDefs that are generally seen as children of this RecordDef. This refers to Records, not the RecordDef ontology, which describes relationships between RecordDefs themselves. e.g., "grid_imaging" Records are often children of "subprojects."
    @attr params Dictionary of all params found in all views (keys), with any default values specified (as value) (read-only attribute).
    @attr paramsK List of all params found in all views (read-only attribute).
    @attr paramsR Parameters that are required for a Record of this RecordDef to validate (read-only attribute).
    @attr owner Current owner of RecordDef. May be different than creator. Gives permission to edit views.
    """

    attr_public = emen2.db.dataobject.BaseDBObject.attr_public | set(["mainview", "views", "private", "typicalchld", "desc_long", "desc_short"])
    attr_required = set(['mainview'])

    # Add custom validators
    _handlers = {}

    def init(self, d):
        super(RecordDef, self).init(d)

        # A string defining the experiment with embedded params
        # this is the primary definition of the contents of the record
        # Required parameter..
        self.__dict__['mainview'] = ''

        # Dictionary of additional (named) views for the record
        self.__dict__['views'] = {}

        # If this is True, this RecordDef may only be retrieved by its owner
        # or by someone with read access to a record of this type
        self.__dict__['private'] = False

        # A list of RecordDef names of typical child records for this RecordDef
        self.__dict__['typicalchld'] = []

        # Short description
        self.__dict__['desc_short'] = None

        # Long description
        self.__dict__['desc_long'] = None

        # The following are automatically generated
        # A dictionary keyed by the names of all params used in any of the views
        # values are the default value for the field.
        # this represents all params that must be defined to have a complete
        # representation of the record. Note, however, that such completeness
        # is NOT REQUIRED to have a valid Record
        self.__dict__['params'] = {}

        # keys from params()
        self.__dict__['paramsK'] = set()

        # required parameters (will throw exception on record commit if empty)
        self.__dict__['paramsR'] = set()


    ##### Setters #####

    def _set_mainview(self, key, value):
        """Only an admin may change the mainview"""
        if not self.isnew() and not self._ctx.checkadmin():
            self.error("Cannot change mainview")

        value = unicode(textwrap.dedent(value))
        ret = self._set('mainview', value, self.isowner())
        self.findparams()
        return ret

    # These require normal record ownership
    def _set_views(self, key, value):
        views = {}
        value = value or {}
        for k,v in value.items():
            views[k] = unicode(textwrap.dedent(v))

        ret = self._set('views', views, self.isowner())
        self.findparams()
        return ret

    def _set_private(self, key, value):
        return self._set('private', int(value), self.isowner())

    # ian: todo: Validate that these are actually valid RecordDefs
    def _set_typicalchld(self, key, value):
        value = emen2.util.listops.check_iterable(value)
        value = filter(None, [unicode(i) for i in value]) or None
        return self._set('typicalchld', value, self.isowner())

    def _set_desc_short(self, key, value):
        return self._set('desc_short', unicode(value or self.name), self.isowner())

    def _set_desc_long(self, key, value):
        return self._set('desc_long', unicode(value or ''), self.isowner())


    ##### RecordDef Methods #####

    # ian: todo: important!! setContext for RecordDef
    # def setContext(self, ctx):
        # def accessible(self):
        #     """Does current Context allow access to this RecordDef?"""
        #     result = False
        #     if not self.private:
        #         result = True
        #     elif self._ctx.username == self.owner:
        #         result = True
        #     elif self._ctx.checkreadadmin():
        #         result = True
        #     return result

    def findparams(self):
        """This will update the list of params by parsing the views"""
        t, d, r = parseparmvalues(self.mainview)

        for i in self.views.values():
            t2, d2, r2 = parseparmvalues(i)
            t |= t2
            r |= r2
            for j in t2:
                # ian: fix for: empty default value in a view unsets default value specified in mainview
                d.setdefault(j, d2.get(j))

        p = {}
        p['params'] = d
        p['paramsK'] = t
        p['paramsR'] = r
        self.__dict__.update(p)

    def validate(self):
        # Run findparams one last time before we commit...
        if not self.mainview:
            self.error("Main protocol (mainview) required")
        self.findparams()

    # This handler implementation is not finalized.
    @classmethod
    def register(cls, names):
        def f(o):
            # print "Registering RecordDef handler:", o
            for name in names:
                if name in cls._handlers:
                    raise ValueError("""RecordDef handler %s already registered""" % name)
                cls._handlers[name] = o
            return o
        return f

    @classmethod
    def get_handler(cls, rec=None):
        handler = None
        if rec:
            handler = rec.get('rectype')
        handler = cls._handlers.get(handler, RecordDefHandler)
        return handler(rec)


# This is not finalized. Subject to change. 
class RecordDefHandler(object):
    def __init__(self, rec):
        self.rec = rec

    def validate(self):
        # print "Base validator...", self.rec
        return




__version__ = "$Revision: 1.42 $".split(":")[1][:-1].strip()



import contextlib
import functools

class CallableGeneratorContextManager(contextlib.GeneratorContextManager):
    '''A subclass of :py:class:`contextlib.GeneratorContextManager` which can be called

    Basically, this exists to avoid unecessary hassles with object methods which need special treatment'''
    def __call__(self, *a, **kw):
        '''Enter self's context and call the object returned by __enter__'''
        with self as obj:
            return obj(*a, **kw)

#NOTE: (ed) this is literally ripped out of contextlib,
# but I've replaced contextlib.GeneratorContextManager with the previous class
def contextmanager(func):
    '''see documentation for :py:class:`contextlib.GeneratorContextManager`'''
    @functools.wraps(func)
    def helper(*args, **kwds):
        return CallableGeneratorContextManager(func(*args, **kwds))
        # ^^^ This is the only line diferent from contextlib.GeneratorContextManager
    return helper


class RegisteredObj(object):
    '''Interface definition for :py:attr:`Registry.child_class`'''
    def __init__(self, name, *args):
        self.name = name

    def update(self, other):
        '''override me'''
        pass

def make_ctxt_manager(name):
    '''create a factory method for making and registering objects suitable to be registered in
    the Registry.

    :param str name: The name of the factory method
    :return: A contextmanager
    '''
    @contextmanager
    def _inner(self, name, *args, **kwargs):
        obj = self.registry.get(name) if name in self.registry else self.child_class(name, *args, **kwargs)
        yield obj
        if obj.name not in self.registry:
            self.register(obj)

    _inner.__name__ = name
    return _inner

import threading
class Registry(object):
    #: The kind of object to be stored in the registry
    #: must, as a minimum implement the :py:class:`RegisteredObj` interface
    child_class = None

    def __setitem__(self, name, value):
        with self._lock:
            self.registry[name] = value

    def __getitem__(self, name):
        with self._lock:
            return self.registry[name]

    def __delitem__(self, name):
        with self._lock:
            del self.registry[name]

    @classmethod
    def reset(cls):
        with cls._lock:
            cls.registry.clear()

    @classmethod
    def get(cls, name, default=None):
        with cls._lock:
            return cls.registry.get(name, default)

    _init_lock = threading.RLock()
    @staticmethod
    def setup(cls):
        '''Decorate the registry with this, this sets the 'registry' attribute to an
        empty dictionary and a factory function for creating child objects.

        The name of the factory function is that of the class the registry registers,
        but lowercase'''

        with cls._init_lock:
            factory_name = cls.child_class.__name__.lower()
            setattr(cls, factory_name, make_ctxt_manager(factory_name))
            cls.registry = {}
            cls._lock = threading.RLock()
            return cls

    def register(self, obj):
        with self._lock:
            old_obj = self.registry.get(obj.name)
            result = obj

            if old_obj is not None:
                old_obj.update(obj)
                result = old_obj
            else:
                self.registry[obj.name] = obj

        return result

### DEMO, read this for a practical example of how the above works:

class DataObj(RegisteredObj):
    def __init__(self, name, a):
        RegisteredObj.__init__(self, name)
        self.a = a
        self.b = None
        self.c = None

@Registry.setup
class DataObjRegistry(Registry):
    child_class = DataObj

if __name__ == '__main__':
    #Demo code

    registry = DataObjRegistry()
    with registry.dataobj('first_obj', 1) as obj1:
        obj1.b = 2

    with registry.dataobj('second_obj', 2) as obj2:
        obj2.b = 3
        obj2.c = 4

    with registry.dataobj('third_obj', 3) as obj3:
        obj1.b = 4

    import random
    def shuffled(lis):
        lis = lis[:]
        random.shuffle(lis)
        return iter(lis)


    for obj_name in shuffled(['first_obj', 'second_obj', 'third_obj']):
        with registry.dataobj(obj_name) as obj:
            print '%s - a: %s, b: %s, c: %s' % (obj.name, obj.a, obj.b, obj.c)



# $Id: proxy.py,v 1.82 2013/05/13 23:33:09 irees Exp $
"""Proxy for accessing EMEN2 API methods

Classes:
    DBProxy
"""

import os
import sys
import time
import collections
import traceback
import weakref
import functools
import inspect

# EMEN2 imports
from emen2.util import listops
import emen2.db.log

##### Warning: This module is very sensitive to changes. #####
##### Please test thoroughly before committing!!         #####

def publicmethod(*args, **kwargs):
    """Decorator for public admin API database method"""
    def _inner(func):
        # print "Registering ", func.func_name
        DBProxy._register_publicmethod(func, *args, **kwargs)
        return func
    return _inner


strht = lambda s, c: s.partition(c)[::2]
def fb():
    return 'hi'

def help(mt):
    def _inner(*a, **b):
        return dict(
            doc = getattr(mt, 'doc', None),
            methods = mt.children.keys()
        )
    return _inner


class MethodTree(object):
    '''Arranges the database methods into a tree so that they can be accessed as db.<a>.<b> (e.g. db.record.get)

    Used by DBProxy.
    '''

    def __init__(self, func=None):
        self.func = func
        if func: self.doc = func.__doc__ or ''
        else: self.doc = ''
        self.children = {}
        self.aliases = {}

    def alias(self, original_name, new_name):
        '''Define an alias for a certain method.

        :param new_name: the name to be replaced
        :param original_name: The replacement name
        '''

        if original_name in self.children:
            raise ValueError, "namespace conflict, cannot alias %r to %r" %(original_name, new_name)
        self.aliases[original_name] = new_name

    def get_alias(self, name):
        '''Check if the method sought has another name.'''
        return self.aliases.get(name,name)

    def __call__(self, *a, **kw):
        return self.func(*a, **kw)

    def __getattr__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)
        name = name.replace('_', '.')
        return self.get_method(name)

    def add_method(self, name, func):
        '''Add a method to the tree

        Overview of the algorithm:

        1. Split the method name at the first '.', the first part is the child to be found at present, the second is the 'tail'
        2. If the child does not exist, add it
        3. If tail is '', then insert 'func' as a method
        4. Otherwise, recurse on the tail

        :param name: the name of the method
        :param func: the function to be executed
        '''
        head, _, tail = name.partition('.') 
        # use partition and not split since it is guaranteed to return a 3-tuple

        self.children.setdefault(head, MethodTree())

        if tail == '':
            self.children[head].func = func
        else:
            self.children[head].add_method(tail, func)

    def get_method(self, name):
        '''Get a method by name, only resolves aliases once

        :param name: the name of the method
        '''
        name = self.get_alias(name)
        head, _, tail = name.partition('.')
        child = self.children.get(head)

        if child is None:
            if tail: raise AttributeError, "method %r not found"%name
            else:
                result = self
                if name == 'help':
                    result = MethodTree(help(self))
        else:
            result = child.get_method(tail)

        return result

class _Method(object):
    # Taken from XML-RPC lib to support nested methods
    def __init__(self, proxy, name):
        self._proxy = proxy
        self._name = name


    def __getattr__(self, name):
        func = self._proxy._publicmethods.get("%s.%s"%(self._name, name))
        if func:
            return self._proxy._wrap(func)
        return _Method(self._proxy, "%s.%s" % (self._name, name))


    def __call__(self, *args):
        raise AttributeError, "No public method %s"%self._name

class DBProxy(object):
    """A proxy that provides access to database public methods and handles low level details, such as Context and transactions.

    db = DBProxy()
    db._login(name, password)
    """

    _publicmethods = {}
    mt = MethodTree()

    @classmethod
    def _allmethods(cls):
        return set(cls._publicmethods)

    def _get_publicmethods(self):
        result = {}
        for x in self._publicmethods:
            cur = result
            for y in x.split('.'):
                ncur = cur.get(y, {})
                if ncur is not cur: cur[y] = ncur
                cur = ncur
        return result

    def __init__(self, db=None, ctx=None, txn=None):
        # it can cause circular imports if this is at the top level of the module
        import database
        db = db or database.DB()

        self._db = db
        self._ctx = ctx
        self._txn = txn

    # Implements "with" interface
    def __enter__(self):
        # print "--> ENTER"
        if self._txn:
            # raise Exception, "DBProxy: Existing open transaction."
            pass
        else:
            self._txn = self._db.dbenv.txncheck(txn=self._txn)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # print "--> EXIT:", exc_type
        if not self._txn:
            raise Exception, "DBProxy: No transaction to close."
        if exc_type is None:
            self._txn = self._db.dbenv.txncommit(txn=self._txn)
        else:
            self._txn = self._db.dbenv.txnabort(txn=self._txn)
        self._txn = None
    
    # Allow to start with write=true/false
    def _newtxn(self, write=False):
        self._txn = self._db.dbenv.txncheck(txn=self._txn, write=write)
        return self
        
    def _gettxn(self):
        return self._txn

    # Rebind a new Context
    def _setContext(self, ctxid=None, host=None):
        self._ctx = self._db._getcontext(ctxid=ctxid, host=host, txn=self._txn)
        self._ctx.setdb(db=self)
        return self

    def _clearcontext(self):
        self._ctx = None

    def _getctx(self):
        return self._ctx

    def _ismethod(self, name):
        if name in self._allmethods():
            return True
        return False

    @classmethod
    def _register_publicmethod(cls, func, apiname=None, write=False, admin=False, ext=False, compat=None):
        apiname = apiname or func.func_name.replace('_','.')
        setattr(func, 'apiname', apiname)
        setattr(func, 'write', write)
        setattr(func, 'admin', admin)
        setattr(func, 'ext', ext)
        setattr(func, 'compat', compat)

        cls._publicmethods[func.apiname] = func
        cls._publicmethods[func.func_name] = func
        if compat:
            cls._publicmethods[compat] = func

        cls.mt.add_method(apiname, func)
        cls.mt.add_method(func.func_name, func)

        if compat:
            cls.mt.add_method(compat, func)

    def _checkwrite(self, method):
        return getattr(self.mt.get_method(method).func, "write", False)

    ##### Wrap DB Public methods #####

    def _callmethod(self, method, args=(), kwargs={}):
        """Call a method by name with args and kwargs (e.g. RPC access)"""
        m = self.mt.get_method(method).func
        if m is not None:
            return self._wrap(m)(*args, **kwargs)

    def __getattr__(self, name):
        if not name.startswith('_'):
            func = self._publicmethods.get(name)
        else:
            func = self.__getattribute__(name)
            
        if func:
            return self._wrap(func)

        return _Method(self, name)

    def _login(self, username, password, host=None):
        ctxid = self.auth.login(username, password)
        self._setContext(ctxid, host)

    def _wrap(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            t = time.time()

            # Remove these from the keyword arguments
            kwargs.pop('ctx', None)
            kwargs.pop('txn', None)

            # Pass the current bound Context
            kwargs['ctx'] = self._ctx
            kwargs['txn'] = self._txn

            emen2.db.log.debug("API: start: %s"%(func.func_name))
            # kwcopy = {}
            # for k,v in kwargs.items():
            #    if k not in ['ctx', 'txn']:
            #        kwcopy[k]=v
            # print "\t<-", args, kwcopy

            if getattr(func, 'admin', False) and not self._ctx.checkadmin():
                raise Exception, "This method requires administrator level access."

            # Make sure the DB is bound to the Context
            self._ctx.setdb(self)

            result = func(self._db, *args, **kwargs)
            ms = (time.time()-t)*1000
            
            emen2.db.log.debug("API: finished: %s in %0.2f ms"%(func.func_name, ms))
            # print "\t->", result
            return result

        return wrapper


__version__ = "$Revision: 1.82 $".split(":")[1][:-1].strip()



# $Id: rss.py,v 1.10 2012/10/18 09:41:40 irees Exp $
from emen2.web.view import View
import time
import datetime
import calendar


class Item(object):
    def __init__(self, title, rec):
        self.title = title
        self.data = rec
        self.name = rec.name
        self.date = rec['creationtime']


@View.register
class RSS(View):
    __matcher__ = dict(
        main=r'^/rss/(?P<begin>\d+)/(?P<end>\d+)/$',
        to_now=r'^/rss/(?P<begin>\d+)/$',
        past_x_time=r'^/rss/-(?P<amount>\d+)(?P<unit>[ymdHMS])/$'
    )

    time_map = {'y': ('Year',  365),
                'm': ('Month', '%m'),
                'd': ('Day',   '%d'),
                'H': ('Hour',  '%H'),
                'M': ('Minute','%M'),
                'S': ('Second','%S')}

    def main(self, begin='', end='', amount='', unit='', *a, **kw):
        self.template = '/pages/rss'
        self.mimetype = 'text/xml; charset=UTF-8'
        if begin != '':
            self.get_data = self.render
            self._begin = time.strftime("%Y/%m/%d %H:%M:%S", time.strptime(begin, '%Y%m%d%H%M%S'))
        # should work, but broken
        #elif amount != '':
        #    self.get_data = self._get_amount
        #    self._amount = int(amount)
        #    self._unit=unit
        if end != '':
            self._end = time.strftime("%Y/%m/%d %H:%M:%S", time.strptime(end, '%Y%m%d%H%M%S'))
        else:
            self._end = None

    def _get_amount(self):
        delta = None
        if self._unit == 'y':
            delta = datetime.timedelta(days=self._amount*365)
        elif self._unit == 'm':
            days = 0
            for c in range(self._amount):
                dys = calendar.mdays[((datetime.date.today().month - c) % 12) + 1]
                # emen2.db.log.info(dys)
                days += dys
            delta = datetime.timedelta(days=days)
        elif self._unit == 'd':
            delta = datetime.timedelta(days=self._amount)
        elif self._unit == 'H':
            delta = datetime.timedelta(seconds=self._amount*3600)
        elif self._unit == 'M':
            delta = datetime.timedelta(seconds=self._amount*60)
        elif self._unit == 'S':
            delta = datetime.timedelta(seconds=self._amount)
        self._begin = datetime.datetime.strptime(self._end, "%Y/%m/%d %H:%M:%S") - delta
        self._begin = self._begin.strftime("%Y/%m/%d %H:%M:%S")
        return self.render()

    def render(self):
        # emen2.db.log.info('begin -> %r' % self._begin)
        # emen2.db.log.info('end -> %r' % self._end)
        query = [['modifytime', 'gt', self._begin]]
        if self._end is not None:
            query.append(['modifytime', 'lte', self._end])
        recs = self.db.query(query)['names']
        recs = self.db.record.get(recs)
        items = []
        for x in recs:
            items.append(Item(x.name, x))
        self.ctxt['items'] = items
        self.title = '%s Record Feed - %s to %s' % (emen2.db.config.get('customization.title'), self._begin, self._end)
        return View.get_data(self)
__version__ = "$Revision: 1.10 $".split(":")[1][:-1].strip()



# $Id: record.py,v 1.92 2013/05/14 08:37:52 irees Exp $
import urllib
import time
import collections

import jsonrpc

import emen2.db.exceptions
import emen2.util.listops as listops
import emen2.web.responsecodes
from emen2.web.view import View

class RecordNotFoundError(emen2.web.responsecodes.NotFoundError):
    title = 'Record not Found'
    msg = 'Record %s not found'



@View.register
class Record(View):
    
    @View.add_matcher(r'^/record/(?P<name>[^/]*)/$')
    def main(self, name=None, children=True, parents=True, sibling=None, viewname="mainview", **kwargs):
        """Main record rendering."""
        # Get record..
        self.rec = self.db.record.get(name, filt=False)
        recnames = self.db.view([name])
        self.title = recnames.get(self.rec.name, self.rec.name)

        # Look for any recorddef-specific template.
        template = '/record/rectypes/%s'%self.rec.rectype
        try:
            emen2.db.config.templates.get_template(template)
        except:
            template = '/record/rectypes/root'
        self.template = template

        # Render main view
        rendered = self.db.view(name, viewname=viewname, options={'output':'html', 'markdown':True})
        rendered_edit = self.db.view(name, viewname=viewname, options={'output':'form', 'markdown':True, 'name': self.rec.name})

        # Some warnings/alerts
        if self.rec.get('deleted'):
            self.notify('Hidden record', error=True)
        if 'publish' in self.rec.get('groups',[]):
            self.notify('Record marked as published data')
        if 'authenticated' in self.rec.get('groups',[]):
            self.notify('Any authenticated user can read this record')
        if 'anon' in self.rec.get('groups', []):
            self.notify('Anyone may access this record anonymously')

        # Find if this record is in the user's bookmarks
        bookmarks = []
        user = None
        try:
            t = time.time()
            user = self.ctxt['USER']
            brec = self.db.record.get(sorted(self.db.rel.children(user.record, rectype='bookmarks')))
            if brec:
                bookmarks = brec[-1].get('bookmarks', [])
        except Exception, e:
            pass
        self.ctxt['bookmarks'] = bookmarks

        # User display names
        users = self.db.user.get([self.rec.get('creator'), self.rec.get('modifyuser')])

        # Parent map
        parentmap = self.routing.execute('Tree/embed', db=self.db, root=self.rec.name, mode='parents', recurse=-1, expandable=False)

        # Children
        children = self.db.record.get(self.rec.children)
        children_groups = collections.defaultdict(set)
        for i in children:
            children_groups[i.rectype].add(i)

        # Siblings
        if sibling == None:
            sibling = self.rec.name
        siblings = self.db.rel.siblings(sibling) # , rectype=self.rec.rectype

        # Get RecordDefs
        recdef = self.db.recorddef.get(self.rec.rectype)
        recdefs = [recdef] + self.db.recorddef.get(children_groups.keys())

        # Pages -- a deprecated UI element. 
        recdefs_d = dict([i.name, i] for i in recdefs)
        pages = collections.OrderedDict()
        pages.uris = {}
        pages['main'] = self.title
        pages.uris
        pages.uris['main'] = self.routing.reverse('Record/main', name=self.rec.name)
        for k,v in children_groups.items():
            pages[k] = "%s (%s)"%(recdefs_d.get(k,dict()).get('desc_short', k),len(v))
            pages.uris[k] = self.routing.reverse('Record/children', name=self.rec.name, childtype=k)    
    
        # Update context
        self.ctxt.update(
            tab = "main",
            rec = self.rec,
            children = children,
            recdef = recdef,
            recdefs = recdefs,
            users = users,
            recnames = recnames,
            parentmap = parentmap,
            edit = False,
            create = self.db.auth.check.create(),
            rendered = rendered,
            rendered_edit = rendered_edit,
            viewname = viewname,
            sibling = sibling,
            siblings = siblings,
            table = "",
            pages = pages
        )    
        
    
    @View.add_matcher(r'^/record/(?P<name>[^/]*)/edit/$', write=True)
    def edit(self, name=None, _format=None, **kwargs):
        self.main(name=name, **kwargs)
        if self.request_method not in ['post', 'put']:
            self.ctxt["tab"] = "edit"
            self.ctxt["edit"] = True
            return

        # Get the record
        if not self.rec.writable():
            raise emen2.db.exceptions.SecurityError, "No write permission for record %s"%self.rec.name

        # Update the record
        if kwargs:
            self.rec.update(kwargs)
            self.rec = self.db.record.put(self.rec)

        for f in self.request_files:
            param = f.get('param', 'file_binary')
            bdo = self.db.binary.put(f)
            self.db.binary.addreference(self.rec.name, param, bdo.name)

        # Redirect
        # IMPORTANT NOTE: Some clients (EMDash) require the _format support below as part of the REST API.        
        self.redirect(self._redirect or self.routing.reverse('Record/main', name=self.rec.name))
        if _format == "json":
            return jsonrpc.jsonutil.encode(self.rec)


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/edit/attachments/$', name='edit/attachments', write=True)
    def edit_attachments(self, name=None, **kwargs):
        self.edit(name=name, **kwargs)
        self.redirect(self.routing.reverse('Record/main', name=self.rec.name, anchor='attachments'))


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/edit/relationships/$', name='edit/relationships', write=True)
    def edit_relationships(self, name=None, **kwargs):
        # ian: todo: Check orphans, show orphan confirmation page
        self.edit(name=name, **kwargs)
        self.redirect(self.routing.reverse('Record/main', name=self.rec.name, anchor='relationships'))


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/edit/permissions/$', name='edit/permissions', write=True)
    def edit_permissions(self, name=None, permissions=None, groups=None, action=None, filt=True):
        self.rec = self.db.record.get(name)
        permissions = permissions or {}
        groups = groups or []
        users = set()
        if hasattr(permissions, 'items'):
            for k,v in permissions.items():
                users |= set(listops.check_iterable(v))
        else:
            for v in permissions:
                users |= set(listops.check_iterable(v))

        if self.request_method != 'post':
            return
            
        if action == 'add':
            self.db.record.setpermissionscompat(names=self.rec.name, recurse=-1, addumask=permissions, addgroups=groups, filt=filt)

        elif action == 'remove':
            self.db.record.setpermissionscompat(names=self.rec.name, recurse=-1, removeusers=users, removegroups=groups, filt=filt)

        elif action == 'overwrite':
            self.db.record.setpermissionscompat(names=self.rec.name, recurse=-1, addumask=permissions, addgroups=groups, filt=filt, overwrite_users=True, overwrite_groups=True)

        else:
            self.rec['groups'] = groups
            self.rec['permissions'] = permissions
            self.rec = self.db.record.put(self.rec)

        self.redirect(self.routing.reverse('Record/main', name=self.rec.name, anchor='permissions'))


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/new/(?P<rectype>[^/]*)/$', write=True)
    def new(self, name=None, rectype=None, _redirect=None, _format=None, **kwargs): 
        """Create a new record."""
        self.main(name=name)
    
        newrec = self.db.record.new(rectype=rectype, inherit=[self.rec.name])
        if self.request_method not in ['post', 'put']:
            self.template = '/record/record.new'
            viewname = 'mainview'
            recdef = self.db.recorddef.get(newrec.rectype)
            rendered = self.db.view(newrec, viewname=viewname)
            self.title = 'New %s'%(recdef.desc_short)
            self.ctxt.update(
                tab = 'new',
                newrec = newrec,
                viewname = viewname,
                rendered = rendered
            )
            return

        # Save the new record
        newrec.update(kwargs)
        newrec = self.db.record.put(newrec)

        for f in self.request_files:
            param = f.get('param', 'file_binary')
            bdo = self.db.binary.put(f)
            self.db.binary.addreference(newrec.name, param, bdo.name)

        # IMPORTANT NOTE: Some clients (EMDash) require the _format support below as part of the REST API.
        self.redirect(_redirect or self.routing.reverse('Record/main', name=newrec.name), content="Your changes were saved.")
        if _format == "json":
            return jsonrpc.jsonutil.encode(newrec)


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/query/$')
    @View.add_matcher(r'^/record/(?P<name>[^/]*)/query/(?P<path>.*)/$')
    def query(self, name=None, path=None, q=None, c=None, **kwargs):
        self.main(name=name)
    

    # @View.add_matcher(r'^/record/(?P<name>[^/]*)/query/(?P<path>.*)/attachments/$')
    @View.add_matcher(r'^/record/(?P<name>[^/]*)/query/attachments/$')
    def query_attachments(self, name=None, path=None, q=None, c=None, **kwargs):
        self.main(name=name)
        self.template = '/record/record.query.attachments'
        # Look up all the binaries
        children = self.db.rel.children(self.rec.name, recurse=-1)
        bdos = self.db.binary.find(record=children, count=0)
        if len(bdos) > 100000 and not confirm:
            raise Exception, "More than 100,000 files returned. Please see the admin if you need to download the complete set."

        records = set([i.record for i in bdos])
        users = set([bdo.get('creator') for bdo in bdos])
        users = self.db.user.get(users)
        # self.ctxt['tab'] = 'attachments'
        self.ctxt['users'].extend(users)
        self.ctxt['recnames'].update(self.db.view(records))
        self.ctxt['bdos'] = bdos
    
    
    @View.add_matcher('^/record/(?P<name>[^/]*)/children/$')
    def children_map(self, name=None):
        self.main(name=name)
        self.template = '/record/record.tree'
        childmap = self.routing.execute('Tree/embed', db=self.db, root=self.rec.name, mode='children', recurse=2, expandable=True, collapse_rectype=["grid_imaging"])
        self.ctxt['childmap'] = childmap


    @View.add_matcher('^/record/(?P<name>[^/]*)/children/(?P<childtype>\w+)/$')
    def children(self, name=None, childtype=None):
        """Main record rendering."""
        self.main(name=name)

        # Child table
        c = [['children', '==', self.rec.name], ['rectype', '==', childtype]]
        query = self.routing.execute('Query/embed', db=self.db, c=c, parent=self.rec.name, rectype=childtype)

        # Update context
        self.ctxt['table'] = query
        self.ctxt['tab'] = 'children-%s'%childtype
        self.ctxt['childtype'] = childtype
        self.ctxt["pages"].active = childtype # Show the active tab -- this might go away at some point.


    @View.add_matcher("^/record/(?P<name>[^/]*)/hide/$", write=True)
    def hide(self, name=None, confirm=False, childaction=None):
        """Main record rendering."""
        self.main(name=name)
        self.template = "/record/record.hide"

        if self.request_method != 'post' or not confirm:
            orphans = self.db.record.findorphans([self.rec.name])
            children = self.db.rel.children(self.rec.name, recurse=-1)
            self.ctxt['confirm'] = confirm
            self.ctxt['orphans'] = orphans
            return

        self.db.record.hide(self.rec.name, childaction=childaction)
        self.redirect(self.routing.reverse('Record/main', name=self.rec.name))



    @View.add_matcher(r'^/record/(?P<name>[^/]*)/history/$')
    @View.add_matcher(r'^/record/(?P<name>[^/]*)/history/(?P<revision>.*)/', name='history/revision')
    def history(self, name=None, simple=False, revision=None):
        """Revision/history/comment viewer"""
        self.main(name=name, parents=True, children=True)
        self.template = "/record/record.history"

        if revision:
            revision = revision.replace("+", " ")

        users = set()
        paramdefs = set()
        users.add(self.rec.get('creator'))
        users.add(self.rec.get('modifyuser'))
        for i in self.rec.get('history',[]) + self.rec.get('comments',[]):
            users.add(i[0])
        for i in self.rec.get('history', []):
            paramdefs.add(i[2])

        users = self.db.user.get(users)
        paramdefs = self.db.paramdef.get(paramdefs)

        # Update context
        # self.ctxt['tab'] = 'history'
        self.ctxt['users'] = users
        self.ctxt['users_d'] = emen2.util.listops.dictbykey(users, 'name') # do this in template
        self.ctxt['paramdefs'] = paramdefs
        self.ctxt['paramdefs_d'] = emen2.util.listops.dictbykey(paramdefs, 'name') # do this in template
        self.ctxt['simple'] = simple
        self.ctxt['revision'] = revision


    @View.add_matcher("^/record/(?P<name>[^/]*)/email/$")
    def email(self, name=None):
        """Email referenced users."""
        self.main(name=name)
        self.template = "/record/record.email"

        pds = self.db.paramdef.find(record=self.rec.keys())
        users_ref = set()
        for i in pds:
            v = self.rec.get(i.name)
            if not v:
                continue
            if i.vartype == "user":
                if i.iter:
                    users_ref |= set(v)
                else:
                    users_ref.add(v)

        users_permissions = set()
        for v in self.rec.get('permissions'):
            users_permissions |= set(v)

        emailusers = self.db.user.get(users_ref | users_permissions)
        self.ctxt['emailusers'] = emailusers


    @View.add_matcher(r'^/record/(?P<name>[^/]*)/publish/$', write=True)
    def publish(self, name=None, state=None):
        self.main(name=name)

        # Get the items
        names = self.db.rel.children(self.rec.name, recurse=-1)
        names.add(self.rec.name)        

        recs = self.db.record.get(names)
        recs_d = emen2.util.listops.dictbykey(recs)
        state = set(map(unicode, state or [])) & names

        # Find published items
        published = set()
        for rec in recs:
            if 'publish' in rec.groups:
                published.add(rec.name)
        self.ctxt['published'] = published
        
        if self.request_method != 'post':
            self.template = '/record/record.publish'
            childmap = self.routing.execute('Tree/embed', db=self.db, root=self.rec.name, mode='children', recurse=2, collapse_rectype='grid_imaging')
            self.ctxt['childmap'] = childmap
            # self.ctxt['tab'] = 'publish'
            return

        # Process form data
        add = state - published
        remove = published - state
        commit = []            
        for i in remove:
            recs_d[i].removegroup('publish')
            commit.append(recs_d[i])
        for i in add:
            recs_d[i].addgroup('publish')
            commit.append(recs_d[i])
            
        recs = self.db.record.put(commit)
        self.redirect(self.routing.reverse('Record/publish', name=self.rec.name))

        



@View.register
class Records(View):
    
    @View.add_matcher(r"^/records/$")
    def main(self, root="0", removerels=None, addrels=None, **kwargs):
        kwargs['recurse'] = kwargs.get('recurse', 2)
        childmap = self.routing.execute('Tree/embed', db=self.db, mode="children", keytype="record", root=root, recurse=kwargs.get('recurse'), id='sitemap')
        self.template = '/pages/records.tree'
        self.title = 'Record relationships'
        self.ctxt['root'] = root
        self.ctxt['childmap'] = childmap
        self.ctxt['create'] = self.db.auth.check.create()


    @View.add_matcher(r"^/records/edit/relationships/$", write=True)
    def edit_relationships(self, root="0", removerels=None, addrels=None, **kwargs):
        self.title = 'Edit record relationships'
        if self.request_method != 'post':
            kwargs['recurse'] = kwargs.get('recurse', 2)
            childmap = self.routing.execute('Tree/embed', db=self.db, mode="children", keytype="record", root=root, recurse=kwargs.get('recurse'), id='sitemap')
            self.template = '/pages/records.tree.edit'
            self.ctxt['root'] = root
            self.ctxt['childmap'] = childmap
            self.ctxt['create'] = self.db.auth.check.create()
            return
            
        self.db.rel.relink(removerels=removerels, addrels=addrels)
        self.redirect('%s/records/edit/relationships/?root=%s'%(self.ctxt.root, root), title=self.title, content="Your changes were saved.")

    @View.add_matcher("^/records/edit/$", write=True)
    def edit(self, *args, **kwargs):
        redirect = kwargs.pop('_redirect', None)
        comments = kwargs.pop('comments', '')
        if self.request_method != 'post':
            return

        for k,v in kwargs.items():
            v['name'] = k

        recs = self.db.record.put(kwargs.values())
        if comments:
            for rec in recs:
                rec.addcomment(comments)
            self.db.record.put(recs)

        if redirect:
            self.redirect(redirect)
            return

        self.simple(content="Saved %s records."%(len(recs)))







__version__ = "$Revision: 1.92 $".split(":")[1][:-1].strip()



# $Id: dataobject.py,v 1.86 2013/05/13 23:33:09 irees Exp $
"""Base classes for EMEN2 DatabaseObjects.

Classes:
    BaseDBObject: Base EMEN2 DBO
    PermissionsDBObject: DBO with additional access controls

"""

import re
import collections
import operator
import hashlib
import UserDict

import emen2.util.listops
import emen2.db.exceptions
import emen2.db.vartypes


class BaseDBObject(object):
    """Base class for EMEN2 DBOs.

    This class implements the mapping interface, and all the required base
    attributes for DBOs:

        name, keytype, creator, creationtime, modifytime, modifyuser, uri

    The name attribute is usually specified by the user when a new item is created,
    but the rest are set by the database when an item is committed. The creator
    and creationtime attributes are set on initial commit, and modifyuser and
    modifytime attributes are usually updated on subsequent commits. The uri
    attribute can be set to indicate an item was imported from an external
    source; presence of the uri attribute will generally mark an item as
    read-only, even to admin users.

    The keytype attribute is set by the BTree storage container.

    The parents and children attributes are valid for classes that allow
    relationships (RelateDB). These are treated specially when an item is
    committed: both the parent and the child will be updated.

    All attributes should also be valid EMEN2 parameters. The default behavior
    for BaseDBObject and subclasses is to validate the attributes as parameters
    when they are set or updated. When a DBO is exported (JSON, XML, etc.) 
    only the attributes listed in cls.attr_public are
    exported. Private attributes may be used by using an underscore
    prefix -- but these WILL NOT BE SAVED, and discarded before committing. An
    example of this behavior is the User._displayname attribute, which
    is recalculated whenever the user is retreived from the database. However,
    the _displayname attribute is still exported by creating a
    displayname class property, and listing that in cls.attr_public. In this way
    it is part of the public interface, even though it is a generated, read-only
    attribute. Another example is the params attribute of Record. This is a
    normal attribute, and read/set from within the class's methods, but is not
    exported directly (it is instead copied into the regular export dictionary.)

    Required attributes can be specified using cls.attr_required. If any of
    these are None, a ValidationError will be raised during commit.

    In addition to implementing the mapping interface, the following methods
    are required as part of the database interface:

        validate         Validate the item before committing
        setContext       Check read permission and bind to a Context
        isowner          Check ownership permission
        isnew            Check if the item has been committed
        writable         Check write permission
        delete           Prepare item for deletion
        rename           Prepare item for renaming

    BaseDBObject also provides the following methods for extending/overriding:

        init             Subclass init
        validate_param   Validate a parameter
        changedparams    Check parameters to re-index
        error            Raise an Exception (default is ValidationError)

    All public methods are safe to override or extend, but be aware of any
    important default behaviors, particularly those related to security and
    validation.

    Naturally, as with anything in Python, anyone with direct
    access to the database can override security by accessing or committing
    to the database with low-level database methods. Therefore, it is generally
    necessary to restrict access using a proxy (e.g. DBProxy) over a network
    connection.

    :attr name: Item name
    :attr creator: Item creator
    :attr creationtime: Creation time, ISO 8601 format
    :attr modifyuser: Last user to modify item
    :attr modifytime: Time of last modification, ISO 8601 format
    :attr uri: Reference to original item if imported

    :classattr attr_public: Public (exported) attributes
    :classattr attr_required: Required attributes
    :property keytype: Key type (default is lower case class name)

    """
    
    attr_public = set(['children', 'parents', 'keytype', 'creator', 'creationtime', 'modifytime', 'modifyuser', 'uri', 'name'])
    attr_protected = set(['creator', 'creationtime', 'modifytime', 'modifyuser', 'uri'])
    attr_required = set()

    def __init__(self, _d=None, **_k):
        """Initialize a new DBO.

        You may provide either a dictionary (first argument or _d keyword)
        or extra keyword arguments.
        """

        # Set the uncommitted flag. This will be stripped out when committed.
        # Check with self.isnew()
        self.__dict__['_new'] = True

        # Copy input and kwargs into one dict
        _d = dict(_d or {})
        _d.update(_k)

        # Temporary setContext
        ctx = _d.pop('ctx', None)
        self.__dict__['_ctx'] = ctx
                
        # Basic attributes
        p = {}
        p['name'] = _d.pop('name', None)
        p['uri'] = _d.pop('uri', None)
        p['keytype'] = _d.pop('keytype', None)
        p['creator'] = ctx.username
        p['creationtime'] = ctx.utcnow
        p['modifyuser'] = ctx.username
        p['modifytime'] = ctx.utcnow
        p['children'] = set()
        p['parents'] = set()
        self.__dict__.update(p)

        # Subclass init
        self.init(_d)

        # Set the context
        self.setContext(ctx)

        # Update with the remaining params
        self.update(_d)


    def init(self, d):
        """Subclass init."""
        pass

    def validate(self):
        """Validate."""
        pass

    def setContext(self, ctx):
        """Set permissions and bind the context."""
        self.__dict__['_ctx'] = ctx

    def changedparams(self, item=None):
        """Differences between two instances."""
        allkeys = set(self.keys() + item.keys())
        return set(filter(lambda k:self.get(k) != item.get(k), allkeys))

    ##### Permissions
    # Two basic permissions are defined: owner and writable
    # PermissiionsDBObject has a more complete permissions model
    # Lack of read access is handled in setContext (raise SecurityError)

    def isowner(self):
        """Check ownership privileges on item."""
        if self._ctx.checkadmin():
            return True
        if self._ctx.username == getattr(self, 'creator', None):
            return True

    def writable(self, key=None):
        """Check write permissions."""
        return self.isowner()

    def isnew(self):
        return getattr(self, '_new', False) == True


    ##### Delete and rename. #####

    def delete(self):
        self.error("No permission to delete.")

    def rename(self):
        self.error("No permission to rename.")

    ##### Required mapping interface #####

    def get(self, key, default=None):
        return self.__getitem__(key, default)

    def has_key(self,key):
        return key in self.attr_public

    def keys(self):
        return list(self.attr_public)

    def items(self):
        return [(k,self[k]) for k in self.keys()]

    def update(self, update):
        """Dict-like update. Returns a set of keys that were updated."""
        cp = set()
        for k,v in update.items():
            cp |= self.__setitem__(k, v)
        return cp

    def _load(self, update):
        """Load from a JSON file; this skips validation on a few keys."""
        if not self.isnew():
            self.error('Cannot update previously committed items this way.')

        # Skip validation for relationships.
        # This will assume they are the correct data format.
        # Note: has to convert to set() here.
        for key in ['parents', 'children']:
           self.__dict__[unicode(key)] = set(update.pop(key, None) or [])

        # Skip validation for defined keys?
        keys = self.attr_public & set(update.keys())
        keys.add('name')
        for key in keys:
            value = update.pop(key, None)
            self.__dict__[unicode(key)] = value

        # Update others...
        # if update:
        #   print "remaining:", update
        #   self.update(update)

    # Behave like dict.get(key) instead of dict[key]
    def __getitem__(self, key, default=None):
        if key in self.attr_public:
            return getattr(self, key, default)
        elif default:
            return default

    def __delitem__(self, key):
        raise AttributeError, 'Key deletion not allowed'

    def __getattr__(self, name):
        return object.__getattribute__(self, name)
        
    # Put everything through setitem for validation/logging/etc..
    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    # Check if there is a method for setting this key,
    # validate the value, set the value, and update the time stamp.
    def __setitem__(self, key, value):
        """Validate and set an attribute or key."""
        # This will validate the parameter, look for a setter, and then call the setter.
        # If a "_set_<key>" method exists, that will always be used for setting.
        # To allow editing of a public attr, there MUST be a _set_<key> method.
        # Then if no setter, and the method is part of the public attrs, then silently return.
        # Finally, use _setoob as the setter. This can allow 'out of bounds' attrs, or raise error (default).
        cp = set()
        if self.get(key) == value:
            return cp

        # Find a setter method (self._set_<key>)
        setter = getattr(self, '_set_%s'%key, None)
        if setter:
            pass
        elif key in self.attr_public:
            # These can't be modified without a setter method defined.
            # (Return quietly instead of SecurityError or KeyError)
            return cp
        else:
            # Setter for parameters that are not explicitly listed as attributes
            # in (attr_public)
            # Default is to raise KeyError or ValidationError
            # Record class will use this to set non-attribute parameters
            setter = self._setoob

        # Validate.
        # *All values* must pass through a validator.
        value = self.validate_param(key, value)

        # The setter might return multiple items that were updated
        # For instance, comments can update other params
        cp |= setter(key, value)

        # Only permissions, groups, and links do not trigger a modifytime update
        if cp - set(['permissions', 'groups', 'parents', 'children']) and not self.isnew():
            self.__dict__['modifytime'] = self._ctx.utcnow
            self.__dict__['modifyuser'] = self._ctx.username
            cp.add('modifytime')
            cp.add('modifyuser')

        # Return all the params that changed
        return cp

    ##### Real updates #####

    def _set(self, key, value, check=None):
        """Actually set a value."""
        # The default permission required to set a key
        # is write permissions. Additionally, the default
        # for write permissions is owners only.
        if check == None:
            check = self.writable()
        if not check:
            msg = "Insufficient permissions to change parameter %s"%key
            self.error(msg, e=emen2.db.exceptions.SecurityError)
        self.__dict__[key] = value
        return set([key])

    def _setoob(self, key, value):
        """Handle params not found in self.attr_public"""
        self.error("Cannot set parameter %s in this way"%key, warning=True)
        return set()

    ##### Update parents / children #####

    def _set_children(self, key, value):
        return self._setrel(key, value)

    def _set_parents(self, key, value):
        return self._setrel(key, value)

    def _setrel(self, key, value):
        """Set a relationship. Make sure we have permissions to edit the relationship."""
        # Filter out changes to permissions on records
        # that we can't access...
        value = set(emen2.util.listops.check_iterable(value))
        orig = self.get(key)
        changed = orig ^ value

        # Get all of the changed items that we can access
        # (KeyErrors will be checked later, during commit..)
        access = self._ctx.db.get(changed, keytype=self.keytype)
        
        # Check write permissions; need write permissions for both.
        for item in access:
            if not (self.writable() and item.writable()):
                msg = 'Insufficient permissions to add or remove relationship: %s -> %s'%(self.name, item.name)
                self.error(msg, e=emen2.db.exceptions.SecurityError)

        # Keep items that we can't access..
        #    they might be new items, or items we won't
        #    have permission to read/edit.
        value |= changed - set(i.name for i in access)
        return self._set(key, value, True)

    ##### Pickle methods #####

    def __getstate__(self):
        """Context and other session-specific information should not be pickled.
        All private keys (starts with underscore) will be removed."""
        odict = self.__dict__.copy() # copy since we are removing keys
        for key in odict.keys():
            if key.startswith('_'):
                odict.pop(key, None)
        return odict

    ##### Validation and error control #####

    # This is the main mechanism for validation.
    def validate_param(self, key, value):
        """Validate a single parameter value."""
        # Check the cache for the param
        hit, pd = self._ctx.cache.check(('paramdef', key))

        # ... otherwise, raise an Exception if the param isn't found.
        if not hit:
            try:
                pd = self._ctx.db.paramdef.get(key, filt=False)
                self._ctx.cache.store(('paramdef', key), pd)
            except KeyError:
                self.error('Parameter %s does not exist' % key)

        # Is it an immutable param?
        if pd.get('immutable') and not self.isnew():
            self.error('Cannot change immutable parameter %s'%pd.name)

        # Validate
        vartype = emen2.db.vartypes.Vartype.get_vartype(pd.vartype, pd=pd, db=self._ctx.db, cache=self._ctx.cache)
        return vartype.validate(value)


    ##### Convenience methods #####

    def error(self, msg='', e=None, warning=False):
        """Raise a ValidationError exception.
        If warning=True, pass the exception, but make a note in the log.
        """
        if e == None:
            e = emen2.db.exceptions.ValidationError
        if not msg:
            msg = e.__doc__
            
        if warning:
            # emen2.db.log.warn("Warning: %s"%e(msg))
            pass
        else:
            raise e(msg)



# A class for dbo's that have detailed ACL permissions.
class PermissionsDBObject(BaseDBObject):
    """DBO with additional access control.

    This class is used for DBOs that require finer grained control
    over reading and writing. For instance, :py:class:`emen2.db.record.Record` 
    and :py:class:`emen2.db.group.Group`. It is a subclass
    of :py:class:`BaseDBObject`; see that class for additional documentation.

    Two additional attributes are provided:
    -    permissions
    -    groups

    The permissions attribute is of the "acl" vartype. It is a list comprised of four
    lists or user names, denoting the following levels of permissions:

    Level 0 - Read
        Permission to read the item

    Level 1 - Comment
        Permission to add comments, if the item supports it

    Level 2 - Write
        Permission to change record attributes/parameters

    Level 3 - Owner
        Permission to change the item's permissions and groups

    The groups attribute is a set of group names. The permissions attribute of
    each group will be overlaid on top of the item's permissions. For instance,
    a user who has comment permissions in a listed group will have comment
    permissions on this item. There are a few built-in groups: administrators,
    read-only administrators, authenticated users, anonymous users, etc. See the
    Group class documentation for additional details.

    Changes to permissions and groups do not trigger an update to the
    modification time and user.

    :attr permissions: Access control list
    :attr groups: Groups

    """
    #These methods are overridden from BaseDBObject:
    #    init, setContext, isowner, writable,
    #The following methods are added to BaseDBObject:
    #    addumask, addgroup, removegroup, removeuser, 
    #     adduser, getlevel, ptest, readable, commentable, 
    #     members, owners, setgroups, setpermissions

    # Changes to permissions and groups, along with parents/children,
    # are not logged.
    attr_public = BaseDBObject.attr_public | set(['permissions', 'groups'])

    def init(self, d):
        """Initialize the permissions and groups

        This method overrides :py:meth:`BaseDBObject.init`
        """
        super(PermissionsDBObject, self).init(d)

        p = {}
        # Results of security test performed when the context is set
        # correspond to, read,comment,write and owner permissions,
        # return from setContext
        p['_ptest'] = [True,True,True,True]

        # Setup the base permissions
        p['permissions'] = [[],[],[],[]]
        p['groups'] = set()
        if self._ctx.username != 'root':
            p['permissions'][3].append(self._ctx.username)

        self.__dict__.update(p)


    ##### Setters #####

    def _set_permissions(self, key, value):
        self.setpermissions(value)
        return set(['_set_groups'])

    def _set_groups(self, key, value):
        self.setgroups(value)
        return set(['groups'])


    ##### Permissions checking #####

    def setContext(self, ctx):
        """Check read permissions and bind Context.

        This method overrides :py:meth:`BaseDBObject.setContext`

        :param ctx: the context to check access against.
        :type: :py:class:`emen2.db.context.Context`
        """

        # Check if we can access this item..
        self.__dict__['_ctx'] = ctx

        # test for owner access in this context.
        if self.isnew() or self._ctx.checkadmin() or self.creator == self._ctx.username:
            self.__dict__['_ptest'] = [True, True, True, True]
            return True

        # Check if we're listed in each level.
        self.__dict__['_ptest'] = [self._ctx.username in level for level in self.permissions]

        # If read admin, set read access.
        if self._ctx.checkreadadmin():
            self._ptest[0] = True
        
        # Apply any group permissions.
        for group in set(self.groups) & self._ctx.groups:
            self._ptest[self._ctx.grouplevels[group]] = True

        # Now, check if we can read.
        if not self.readable():
            raise emen2.db.exceptions.SecurityError, "Permission denied: %s %s"%(self.keytype, self.name)
        return True

    def getlevel(self, user):
        """Get the user's permissions for this object

        :rtype: int
        """
        for level in range(3, -1, -1):
            if user in self.permissions[level]:
                return level

    def isowner(self):
        """Is the current user the owner?

        This method overrides :py:meth:`BaseDBObject.isowner`

        :rtype: bool
        """
        return self._ptest[3]

    def readable(self):
        """Does the user have permission to read the record(level 0)?

        :rtype: bool
        """
        return any(self._ptest)

    def commentable(self):
        """Does user have permission to comment (level 1)?

        :rtype: bool
        """
        return any(self._ptest[1:])

    def writable(self):
        """Does the user have permission to change the record (level 2)?

        This method overrides :py:meth:`BaseDBObject.writable`

        :rtype: bool
        """
        return any(self._ptest[2:])

    def members(self):
        """Get all users with read permissions.

        :rtype: [str]
        """
        ###TODO:documentation: what does this do?
        return set(reduce(operator.concat, self.permissions))

    def owners(self):
        """Get all users with ownership permissions.

        :rtype: [str]
        """
        return self.permissions[3]

    def ptest(self):
        """Get a tuple with permission checks for each level"""
        return self._ptest

    ##### Users #####

    def _check_permformat(self, value):
        if hasattr(value, 'items'):
            v = [[],[],[],[]]
            ci = emen2.util.listops.check_iterable
            v[0] = ci(value.get('read'))
            v[1] = ci(value.get('comment'))
            v[2] = ci(value.get('write'))
            v[3] = ci(value.get('admin'))
            value = v
        permissions = [[unicode(y) for y in x] for x in value]
        if len(permissions) != 4:
            raise ValueError, "Invalid permissions format"
        return permissions

    def adduser(self, users, level=0, reassign=False):
        """Add a user to the record's permissions

        :param users: A list of users to be added to the permissions
        :type users: [str]
        :param level: The permission level to give to the users
        :type level: int
        :param reassign: Whether or not the users added should be reassigned. (default False)
        :type reassign: bool
        """
        if not users:
            return

        if not hasattr(users,"__iter__"):
            users = [users]

        level = int(level)
        if not 0 <= level <= 3:
            raise Exception, "Invalid permissions level. 0 = Read, 1 = Comment, 2 = Write, 3 = Owner"

        p = [set(x) for x in self.permissions]

        # Root is implicit
        users = set(users) - set(['root'])

        if reassign:
            p = [i-users for i in p]

        p[level] |= users

        p[0] -= p[1] | p[2] | p[3]
        p[1] -= p[2] | p[3]
        p[2] -= p[3]

        self.setpermissions(p)

    def addumask(self, value, reassign=False):
        """Set permissions for users in several different levels at once.

        :param value: The list of users
        :type value: [ [str], [str], [str] ]
        :param reassign: Whether or not the users added should be reassigned. (default False)
        :type reassign: bool

        """
        umask = self._check_permformat(value)

        p = [set(x) for x in self.permissions]
        umask = [set(x) for x in umask]
        users = reduce(set.union, umask)

        if reassign:
            p = [i-users for i in p ]

        p = [j|k for j,k in zip(p,umask)]
        p[0] -= p[1] | p[2] | p[3]
        p[1] -= p[2] | p[3]
        p[2] -= p[3]

        self.setpermissions(p)

    def removeuser(self, users):
        """Remove a user from permissions"""
        if not users:
            return

        p = [set(x) for x in self.permissions]
        if not hasattr(users, "__iter__"):
            users = [users]
        users = set(users)
        p = [i-users for i in p]

        self.setpermissions(p)

    def setpermissions(self, value):
        """Set the permissions"""
        value = self._check_permformat(value)
        return self._set('permissions', value, self.isowner())

    ##### Groups #####

    def addgroup(self, groups):
        """Add a group to the record"""
        if not hasattr(groups, "__iter__"):
            groups = [groups]
        g = self.groups | set(groups)
        self.setgroups(g)

    def removegroup(self, groups):
        """Remove a group from the record"""
        if not hasattr(groups, "__iter__"):
            groups = [groups]
        g = self.groups - set(groups)
        self.setgroups(g)

    def setgroups(self, groups):
        """Set the object's groups"""
        groups = emen2.util.listops.check_iterable(groups)
        return self._set('groups', set(groups), self.isowner())



__version__ = "$Revision: 1.86 $".split(":")[1][:-1].strip()



# $Id: log.py,v 1.8 2012/10/18 09:41:40 irees Exp $
# Standard View imports
import functools
import collections
import os.path

import emen2.db.config
from emen2.web.view import View, AdminView
from emen2.util.loganalyzer import AccessLogFile, AccessLogLine
import datetime


class _Null: pass
def cast_arguments(*postypes, **kwtypes):
    def _func(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            out = []
            for typ, arg in itertools.izip_longest(postypes, args, fillvalue=_Null):
                if arg != _Null:
                    if typ != _Null and typ != None:
                        arg = typ(arg)
                    out.append(arg)
            for k,v in kwargs.iteritems():
                typ = kwtypes.get(k, _Null)
                if typ != _Null and typ != None:
                    kwargs[k] = typ(kwargs[k])
            return func(*args, **kwargs)
        return _inner
    return _func



class RecordNotFound(emen2.web.responsecodes.NotFoundError):
    title = 'Record Not Found'
    msg = 'Record %d not found'
    def __init__(self, msg):
        emen2.web.responsecodes.NotFoundError.__init__(self, msg)

class Reverser(object):
    def __init__(self, fil):
        self._file = fil
        self._file.seek(-1, 2)
        self.stopped = False
    def _bseek(self):
        result = False
        if self._file.tell() == 1: result = True
        else:
            self._file.seek(-2,1)
        self.stopped = result
        return result
    def next(self):
            if self.stopped: raise StopIteration
            line = []
            line.insert(0, self._file.read(1))
            if line[0] == '\n':
                line.pop()
            self._bseek()
            if not self.stopped:
                cur = self._file.read(1)
                while cur != '\n' and not self.stopped:
                    line.insert(0, cur)
                    self._bseek()
                    cur = self._file.read(1)
                self._bseek()
            return ''.join(line)
    def __iter__(self):
        try:
            while True:
                yield self.next()
        except EOFError: raise StopIteration

# class TableJS(emen2.web.templating.BaseJS):
#     def main(self):
#         super(TableJS, self).init()
#         self.files = '%s/static/js/jquery/jquery.dataTables.js' % g.ROOT
#         self.files = '%s/static/js/jquery/jquery.accordion.js' % g.ROOT

ident = lambda x:x
class Normalizer(object):
    def __init__(self, norm=ident, unnorm=ident):
        self.norm = norm
        self.unnorm = unnorm

def decorator(func):
    @functools.wraps(func)
    def _inner(func1):
        @functools.wraps(func1)
        def _inner1(*a, **kw):
            return func(func1(*a, **kw))
        return _inner1
    return _inner

@decorator
def stringify(dict_):
    sdict = {'TB': (0, '{TB} TB'), 'GB': (1, '{GB} GB'), 'MB': (2, '{MB} MB'), 'KB': (3,'{KB} KB'), 'B':(4,'{B} B')}
    fmtl =  [ x[1] for x in sorted( sdict[k] for k in (set(dict_) & set(sdict)) ) ]
    return ', '.join(fmtl).format(**dict_)


@stringify
def makereadable(val):
    vals = dict(B = val, KB = 0, MB = 0, GB = 0, TB = 0)
    def adjust(dict_, a,b):
        if dict_[a] >= 1024:
            dict_[b] = dict_[a] / 1024
            dict_[a] %= 1024
    adjust(vals, 'B', 'KB')
    adjust(vals, 'KB', 'MB')
    adjust(vals, 'MB', 'GB')
    adjust(vals, 'GB', 'TB')
    vals = dict( (k,v) for k,v in vals.iteritems() if v != 0 ) or {'B':0}
    return vals

@View.register
@AdminView.attach
class LogAnalysis(View):#AdminView):
    template = '/pages/log'
    
    @View.add_matcher("^/log/main/(?P<dataset>.*)/$")
    def main(self, dataset='host', start=0, end=100, sort='host', reverse=1):
        pass

    def _getlines(self, file_, logclass, start, end, reverse=True, index=None):
        lines, errors = [],[]

        logpath = emen2.db.config.get('paths.log')
        with file(os.path.join(logpath, file_)) as f:
            extra = lambda x:x
            if reverse: extra = Reverser

            for n, line in enumerate( x for x in extra(f) if x.strip() ): #filter out blank lines and return n,line = ( line #, line )
                try:
                    if n >= end: break
                    elif start <= n:
                        lines.append(line)
                except Exception, e:
                    emen2.db.log.error(e)
                    errors.append('problem, line: %r' % line)

        logfile = logclass(*lines, index=index)
        return logfile, errors

    @View.add_matcher("^/log/access/(?P<dataset>.*)/$")
    @cast_arguments(start=int, end=int, reverse=lambda x: bool(int(x)), dataset=lambda x: x.split('/'))
    def access(self, dataset='host', start=0, end=100, sort='host', reverse=True):
        name = dataset[0]
        index = set([name])
        linefilter = collections.defaultdict(set)

        jsfilter = []
        data_cast = dict(AccessLogLine.order)
        for x in dataset[1:]:
            n = x.split(':', 1)
            print n
            if (len(n) == 2) and (n[0] in AccessLogLine.labels):
                label, val = n
                index.add(label)
                linefilter[label].add(data_cast.get(label, ident)(val))
            else:
                jsfilter.append(x)

        ctx = self.db._getctx()
        if not ctx.checkadmin():
            linefilter['username'] = set([ctx.username])

        logfile, errors = self._getlines('access.log', AccessLogFile, start, end, reverse=reverse, index=index)

        data = ldata = logfile.data.get(name, {})
        if linefilter:
            print 1
            data = {}
            for name, lines in ldata.iteritems():
                data[name] = []
                for line in lines:
                    to_check = set(line) & set(linefilter)
                    add = True

                    for key in to_check:
                        val = line[key]
                        if val not in linefilter[key]:
                            add = False

                    if add:
                        data[name].append(line)
                if len(data[name]) == 0: del data[name]



        print linefilter, jsfilter
        self.ctxt.update(
            sort=sort, errors=errors, name=name,

            title='Access Log',
            dataset=data,
            filter_str=' '.join(jsfilter),

            viewargs=dict(
                start=start, end=end, sort=sort, reverse=int(reverse)
            ),

            norm = dict(
                    rtime = Normalizer(
                        lambda x: (x and x.strftime('%Y%m%d%H')),
                        lambda x: (x and datetime.datetime.strptime(x, '%Y%m%d%H'))
                    ),
                    size = Normalizer(
                        lambda x: x/(1024),
                        lambda x: ('%s-%s' % (makereadable(x*(1024)), makereadable(x*1024 + 1023)))
                    )
                ).get(name, Normalizer())
        )
__version__ = "$Revision: 1.8 $".split(":")[1][:-1].strip()
